body {
    background: #eeeeee;
    margin: 2em;
    text-align: center;
    font-family: system-ui, sans-serif;
}

footer {
    padding: 2em;
    text-align: center;
    height: 10vh;
}

.ui.segment {
    background: #eeeeee;
}

#graphs {
    display: grid;
    width: 100%;
    height: 300px;
    grid-template-columns: 1fr 1fr;
}
#graphs_center {
    display: block;
    margin-top: 10px;
    margin-bottom: 10px;
    width: 100%;
    height: 300px;
}
#left {
    margin-right: 2px;
    height: 300px;
}
#right {
    height: 300px;
}
#center {
    height: 300px;
}
#packetsChart, #messageChart, #emailChart, #memChart {
    border: 1px solid #ccc;
    background: #ddd;
}
#stats {
    margin: auto;
    width: 80%;
}
#jsonstats {
    display: none;
}
#title {
    font-size: 4em;
}
#version{
    font-size: .5em;
}
#uptime, #aprsis {
    font-size: 1em;
}
#callsign {
    font-size: 1.4em;
    color: #00F;
    padding-top: 8px;
    margin:10px;
}

#title_rx {
    background-color: darkseagreen;
    text-align: left;
}

#title_tx {
    background-color: lightcoral;
    text-align: left;
}

.aprsd_1 {
    background-image: url(/static/images/aprs-symbols-16-0.png);
    background-repeat: no-repeat;
    background-position: -160px -48px;
    width: 16px;
    height: 16px;
}
